/*******************************************************************************
* Licensed Materials - Property of HCL
* (c) Copyright HCL Technologies Ltd. 2017, 2019.  All Rights Reserved.
*
* Note to U.S. Government Users Restricted Rights:
* Use, duplication or disclosure restricted by GSA ADP Schedule
* Contract with IBM Corp.
*******************************************************************************/

/**
 * Deploy the LIVE server.
 * - start Mongo DB (unless it's already running)
 * - start the LIVE server using PM2
 *
 * @author Mattias Mohlin
 */

'use strict';

let path = require('path');
let conf = require('nconf');
require('nconf-strip-json-comments')(conf);
let mongo = require('mongodb');
let fs = require('fs');
let shell = require('shelljs');
let term = require('terminal-kit').terminal;
let tmp = require('tmp');
let pm2 ;

conf.file('default_config', {
    file: './dr-config.json'
});

// If the DR server is running it will be stopped.
function stopDRServerIfRunning(pm2ConfigFile,callback) {
	
	// Stop by deleting the app from PM2 (in case a new version of the app is about to be deployed)
        pm2.delete(pm2ConfigFile, (err) => {
            if (err) {
                console.log('Failed to stop LIVE servers. Make sure PM2 is installed by running install-dr first.');
                console.log(err);
                process.exit(1);
            }

            console.log('Successfully stopped LIVE servers!');
            callback();
        });
 
	
}

// Start LIVE server in a mode where it just reports the configuration settings it will use.
// The configuration settings are provided as the first argument of the callback function.
function getDRServerConfig(pm2ConfigFile, callback) {
	// Connect to an existing PM2 daemon or launch a new one
    pm2.connect((err) => {
        if (err) {
            console.log('Failed to connect to PM2: ' + err);
            console.log('Make sure PM2 is installed by running install-dr first.');
            process.exit(1);
        }

		callback();
    });
}

// Start LIVE server
function startDRServer(pm2ConfigFile, callback) {
    // Connect to an existing PM2 daemon or launch a new one
    pm2.connect((err) => {
        if (err) {
            console.log('Failed to connect to PM2: ' + err);
            console.log('Make sure PM2 is installed by running install-dr first.');
            process.exit(1);
        }

        // Stop the DR server if it's already running
        stopDRServerIfRunning(pm2ConfigFile,() => {
            console.log('Starting LIVE servers now using settings from ' + pm2ConfigFile);

            pm2.start(pm2ConfigFile, (err, proc) => {
                if (err) {
                    console.log('Failed to start LIVE server.');
                    console.log(err);
                    process.exit(1);
                }

                console.log('Successfully started LIVE server!');
	       //Temporarily disabling it
               // console.log('LIVE server status is now: ' + proc[0].pm2_env.status);

                console.log('Server URL is: https://localhost/home.html');// + serverConfig.dr_host + ':' + serverConfig.dr_port + '/index.html');

                callback();
            });
        });
    });
}

// Attempt to connect to Mongo
function startMongoIfNeeded(mongoURL, mongoExe, dataFolder, logfile, callback) {
    let client = mongo.MongoClient;
    client.connect(mongoURL, function (err, db) {
        if (err) {
            console.log('Could not detect a running Mongo database at ' + mongoURL + '. Start it now...');

            if (!fs.existsSync(mongoExe)) {
                console.log('Failed to find ' + mongoExe + '! Did you forget to run install-dr?');
                process.exit(1);
            }

            if (!fs.existsSync(dataFolder)) {
                // The data folder must exist before we can start Mongo. Create it now.
                console.log('Folder ' + dataFolder + ' is missing. Creating it now.');
                shell.mkdir('-p', dataFolder);
                if (shell.error()) {
                    console.log('Failed to create folder ' + dataFolder);
                    process.exit(1);
                }
            }

            let out = fs.openSync(logfile, 'a');
            let err = fs.openSync(logfile, 'a');

            let spawn = require('child_process').spawn;

            let childProcess = spawn(mongoExe, ['--dbpath', dataFolder], { detached: true, stdio: ['ignore', out, err] });
            childProcess.unref();

            childProcess.on('error', function (err) {
                console.log('Failed to start Mongo!');
                process.exit(1);
            });

            childProcess.on('close', function (code) {
                console.log('Mongo terminated unexpectedly!');
                process.exit(1);
            });

            // Wait until Mongo is up and running            
            function waitUntilStarted(attempts) {
                if (attempts >= conf.get('mongo_max_startup_time')) {
                    console.log('Mongo failed to start within ' + conf.get('mongo_max_startup_time') + ' seconds!');
                    process.exit(1);
                }
                setTimeout(() => {
                    client.connect(mongoURL, function (err, db) {
                        if (err) {
                            console.log('Waited ' + attempts + ' of ' + conf.get('mongo_max_startup_time') + ' seconds...');
                            waitUntilStarted(attempts + 1);
                            return;
                        }
                        console.log('Detected a running Mongo database after ' + attempts + ' seconds!');
                        console.log('See mongostartup.log for details.');
                        callback();
                    });
                }, 1000);
            }
            waitUntilStarted(1);
        }
        else {
            console.log('Detected a running Mongo database, so it will not be started again.');
            callback();
        }
    });
}

function linkPm2(callback) {
    console.log("Creating link for pm2 installed globally");
    let exec  = require('child_process').exec;
    let child = exec('npm link pm2', function(error, stdout, stderr) {
        if (error !== null) {
            console.log('Create pm2 link failed: ' + error);
            process.exit(1);
        } else {
            console.log('pm2 link created');
            callback();
        }
    });
}
if (process.platform == 'win32') {
    var mongoExe = conf.get('mongo_win_install_folder') + 'bin\\mongod.exe';
    var dataFolder = conf.get('mongo_win_data_folder');
}
else if (process.platform == 'linux') {
    var mongoExe = conf.get('mongo_linux_install_folder') + 'bin/mongod';
    var dataFolder = conf.get('mongo_linux_data_folder');
}

let mongoLogfile = conf.get('mongo_log_file');
let pm2ConfigFile = conf.get('pm2_config_file');

// As the very first step, launch the LIVE server in a mode where it just reports which configuration settings it will use.
// We need to know these real settings, rather than what is obtained from ../OnPrem_Design_Room/conf since the PM2 config file
// may override some of the settings.
linkPm2(() => {
 pm2 = require('pm2');
getDRServerConfig(pm2ConfigFile,() => {

    const mongoURL = 'mongodb://localhost:27017/';
	
    // Start Mongo (if needed)
    startMongoIfNeeded(mongoURL, mongoExe, dataFolder, mongoLogfile, () => {

        // Start LIVE server using PM2
        startDRServer(pm2ConfigFile, () => {
            console.log('All done!');
            process.exit(0);
            });
        });

    });

});