/*******************************************************************************
* Licensed Materials - Property of HCL
* (c) Copyright HCL Technologies Ltd. 2017.  All Rights Reserved.
*
* Note to U.S. Government Users Restricted Rights:
* Use, duplication or disclosure restricted by GSA ADP Schedule
* Contract with IBM Corp.
*******************************************************************************/

/**
 * Install the LIVE server.
 * - install Mongo DB
 * - install PM2
 *
 * @author Mattias Mohlin
 */

'use strict';

let conf = require('nconf');
require('nconf-strip-json-comments')(conf);
let path = require('path'); 
let fs = require('fs');
let term = require('terminal-kit').terminal;

conf.argv().env();
let configFile = conf.get('DR_CONFIG_FILE');
if (configFile) {
    if (!fs.existsSync(configFile)) {
        console.log('Failed to find configuration file specified by DR_CONFIG_FILE: ' + configFile);
        process.exit(1);
    }
    conf.file('custom_config', configFile);
}
conf.file('default_config', {
    file: './dr-config.json'
});

// Installs Mongo DB
function installMongo(name, file, targetFolder) {
    return new Promise((resolve, reject) => {            
        term('Do you want to install ' + name + ' into ' + targetFolder + '? [Y|n]\n');
        term.yesOrNo({ yes: ['y', 'Y', 'ENTER'], no: ['n', 'N'] }, function (error, result) {
            if (!result) {
                console.log('Skipping installation of ' + name);
                resolve();
                return;
            }
            console.log('Installing ' + name + ' from ' + file + ' into ' + targetFolder);    

            if (!fs.existsSync(file)) {
                console.log('Mongo installation file ' + file + ' does not exist!');
                process.exit(1);
            }

            let exec = require('child_process').exec;
            // Mongo is only installed automatically on Windows...
            let cmd = 'call msiexec.exe /q /i ' + file + ' /l*vx mongoinstall.log INSTALLLOCATION="' + targetFolder + '" ADDLOCAL="Server,ServerService,Client,MonitoringTools,ImportExportTools,MiscellaneousTools" SHOULD_INSTALL_COMPASS="0"';            
            
            exec(cmd, (error, stdout, stderr) => {
                console.log(stdout);
                if (error) {
                    console.log('Failed to install ' + name + '!');
                    console.log('Installer terminated with code: ' + error.code);
                    process.exit(1);
                }
                console.log('Successfully installed ' + name + ' into ' + targetFolder + '! See mongoinstall.log for details.');
                resolve();
            });
        });
    });        
}

// Installs an npm package from a local npmbox file
function npmunbox(pkg, file) {
    return new Promise((resolve, reject) => {        
        term('Do you want to install ' + pkg + '? [Y|n]\n');
        term.yesOrNo({ yes: ['y', 'Y', 'ENTER'], no: ['n', 'N'] }, function (error, result) {
            if (!result) {
                console.log('Skipping installation of ' + pkg);
                resolve();
                return;
            }

            console.log('Installing ' + pkg + ' from ' + file + '...');

            let npmunboxLocation = path.join(__dirname, 'node_modules', 'npmbox', 'bin', 'npmunbox');
            let spawn = require('child_process').spawn;

            let childProcess = spawn('node', [npmunboxLocation, '-g', file]);

            childProcess.on('error', function (err) {
                console.log('Failed to install ' + pkg + '!');
                process.exit(1);
            });

            childProcess.stdout.on('data', function (data) {
                console.log('' + data);
            });

            childProcess.stderr.on('data', function (data) {
                console.log('' + data);
            });

            childProcess.on('close', function (code) {
                if (code != 0) {
                    console.log('Process of unboxing ' + file + ' terminated unexpectedly!');
                    process.exit(1);
                }
                console.log('Successfully installed ' + pkg + '!');
                resolve();
            });
        });        
    });
}

if (process.platform == 'win32') {
    // Install Mongo on Windows
    term('Mongo DB is licensed under GNU AGPL v3.0 (https://www.mongodb.com/community/licensing).\n');    
    var mongoInstall = installMongo('Mongo (Windows)', 'downloaded\\mongo-win.msi', conf.get('mongo_win_install_folder'));
}
else if (process.platform == 'linux') {
    // Do not install Mongo on Linux
    var mongoInstall = Promise.resolve(false);
}
else {
    console.log('Unsupported platform. LIVE can only be installed on Windows and Linux systems.');
    process.exit(1);
}

mongoInstall.then(() => {
    term('PM2 is licensed under GNU AGPL v3.0 (https://github.com/Unitech/pm2/blob/master/GNU-AGPL-3.0.txt).\n');        
    return npmunbox('pm2', 'downloaded/pm2.npmbox');  
})
.then(() => {
    if (process.platform == 'linux') {
        term('NOTE: Mongo DB needs to be manually installed on Linux systems. Follow instructions for your Linux distribution at ');
        term.inverse('https://docs.mongodb.com/manual/administration/install-on-linux/#recommended-linux-install'); term('\n');
    }
    term('All done! You may now start LIVE by running '); term.inverse('node deploy-dr.js'); term('.\n');
    process.exit(0);
});


